(function(mod) {
	if (typeof exports == 'object' && typeof module == 'object') // CommonJS
		mod(require('codemirror'));
	else if (typeof define == 'function' && define.amd) // AMD
		define(['codemirror'], mod);
	else // Plain browser env
		mod(CodeMirror);
})(function(CodeMirror) {

	const commonKeywords = ['if', 'for', 'while', 'fun', 'else', 'with', 'return', 'break', 'continue', 'class', 'import', 'and', 'or'];
	const commonBuiltins = ['len', 'num', 'time', 'str', 'clock', 'input', 'random', 'pow', 'bool', 'floor', 'ceil', 'print', 'println', 'exit'];

	CodeMirror.registerHelper('hintWords', 'caboose', commonKeywords.concat(commonBuiltins));

	const wordRegexp = (words) => {
		return new RegExp('^((' + words.join(')|(') + '))\\b');
	}

	class Context {
		constructor(indented, column, type, align, prev) {
			this.indented = indented;
			this.column = column;
			this.type = type;
			this.align = align;
			this.prev = prev;
		}
	}

	CodeMirror.defineMode('caboose', (conf) => {
		const indentUnit = conf.indentUnit;
		const ERROR = 'error';

		const keywords = wordRegexp(commonKeywords);
		const builtins = wordRegexp(commonBuiltins);
		const brackets = /[\(\)\[\]\{\}]/g;
		const atoms = wordRegexp(['true', 'false', 'nil', 'this']);
		const operatorChars = /([<=!>+-/*%])/g;

		let curPunc;

		const tokenString = (quote) => {
			return function(stream, state) {
				let escaped = false, next, end = false;
				while ((next = stream.next()) != null) {
					if (next == quote && !escaped) {
						end = true;
						break;
					}
					escaped = !escaped && next == "\\";
				}
				if (end || !escaped) state.tokenize = tokenBase;
				return "string";
			};
		}

		const pushContext = (state, col, type) => {
			return state.context = new Context(state.indented, col, type, null, state.context);
		};

		const popContext = (state) => {
			if (!state.context.prev) return;
			const t = state.context.type;
			if (t == ')' || t == ']' || t == '}') state.indented = state.context.indented;
			return state.context = state.context.prev;
		};

		const tokenComment = (stream, state) => {
			let maybeEnd = false, ch;
			while(ch = stream.next()) {
				if (ch == '/' && maybeEnd) {
					state.tokenize = tokenBase;
					break;
				}

				maybeEnd = (ch == '*');
			}
		};

		const tokenBase = (stream, state) => {
			const ch = stream.next();
			if (ch == '"' || ch == '\'') {
				state.tokenize = tokenString(ch);
				return state.tokenize(stream, state);
			}

			if (/[\d\.]/.test(ch)) {
				if (ch == '.') stream.match(/^[0-9]+([\-+]?[0-9]+)?/);
				else stream.match(/^[0-9]*\.?[0-9]*([\-+]?[0-9]+)?/);
				return 'number';
            }

            if (/[\[\]\{\}\(\)]/.test(ch)) {
                curPunc = ch;
                return 'bracket'
            }
            
			if (/[,;\:\.]/.test(ch)) {
				curPunc = ch;
				return null;
            }
            
			if (ch == '/') {
				if (stream.eat('*')) {
					state.tokenize = tokenComment;
					return tokenComment(stream, state);
				}
				if (stream.eat('/')) {
					stream.skipToEnd();
					return 'comment';
				}
			}

			if (operatorChars.test(ch)) {
				stream.eatWhile(operatorChars);
				return 'operator';
			}

			stream.eatWhile(/[\w](?:[\w\d])*/);
			var cur = stream.current();
            if (keywords.includes(cur)) return 'keyword';
            if (builtins.includes(cur)) return 'builtin'
			if (atoms.includes(cur)) return 'atom';
			return 'variable';
		}

		return {
			startState: (baseColumn) => {
				return {
					tokenize: null,
					context: new Context((baseColumn || 0) - indentUnit, 0, 'top', false),
					indented: 0,
					startOfLine: true,
				};
			},

			token: (stream, state) => {
				const ctx = state.context;

				if (stream.sol()) {
					if (ctx.align == null) ctx.align = false;
					state.indented = stream.indentation();
					state.startOfLine = true;
				}

				if (stream.eatSpace()) return null;
				curPunc = null;
				
				const style = (state.tokenize || tokenBase)(stream, state);
				if (style == 'comment') return style;
				if (ctx.align == null) ctx.align = true;

				if (curPunc == '{') pushContext(state, stream.column(), '}');
				else if (curPunc == '[') pushContext(state, stream.column(), ']');
				else if (curPunc == '(') pushContext(state, stream.column(), ')');
				else if (curPunc == ctx.type) popContext(state);
				state.startOfLine = false;
				return style;
			},

			indent: (state, textAfter) => {
				if (state.tokenize != tokenBase && state.tokenize != null) return CodeMirror.Pass;
				const ctx = state.context, firstChar = textAfter && textAfter.charAt(0);
				const closing = firstChar == ctx.type;
				if (ctx.align) return ctx.column + (closing ? 0 : 1);
				else return ctx.indented + (closing ? 0 : indentUnit);
			},

			fold: 'brace',
			blockCommentStart: '/*',
			blockCommentEnd: '*/',
			lineComment: '//',
			closeBrackets: '()[]{}\'\'""',
		};
    });

    CodeMirror.defineMIME('text/x-caboose', 'caboose');
});